/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "pfs_sysdeps.h"
#include "pfs_channel.h"
#include "pfs_channel_cache.h"
#include "pfs_process.h"
#include "pfs_sys.h"
#include "pfs_poll.h"
#include "pfs_service.h"

extern "C" {
#include "tracer.h"
#include "stringtools.h"
#include "full_io.h"
#include "xmalloc.h"
#include "int_sizes.h"
#include "macros.h"
#include "debug.h"
}

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <limits.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <sys/file.h>
#include <termios.h>
#include <sys/time.h>
#include <ctype.h>
#include <sys/utsname.h>
#include <sys/un.h>
#include <net/if.h>
#include <linux/sockios.h>

extern struct pfs_process *current;
extern char *pfs_temp_dir;
extern char *pfs_false_uname;
extern uid_t pfs_uid;
extern gid_t pfs_gid;

extern pid_t trace_this_pid;

extern long long pfs_syscall_count;
extern long long pfs_read_count;
extern long long pfs_write_count;

extern int pfs_trap_after_fork;

extern void handle_specific_process( pid_t pid );

/*
Divert this incoming system call to a read or write on the I/O channel
*/

void divert_to_channel( struct pfs_process *p, int syscall, const void *uaddr, int length, pfs_size_t channel_offset )
{
	int args[5];
	args[0] = pfs_channel_fd();
	args[1] = (int) uaddr;
	args[2] = length;
	args[3] = (int) (channel_offset);
	args[4] = (int) (((UINT64_T)channel_offset) >> 32);
	tracer_args_set(p->tracer,syscall,args,5);
	p->syscall_args_changed = 1;
	p->diverted_length = length;
}

/*
Divert this incoming system call to something harmless with the given result.
*/

void divert_to_dummy( struct pfs_process *p, int result )
{
	p->syscall_dummy = 1;
	p->syscall_result = result;
	tracer_args_set(p->tracer,SYSCALL_getpid,0,0);
}

int errno_in_progress( int e )
{
	return (e==EAGAIN || e==EALREADY | e==EINPROGRESS);
}

/*
The large majority of calls, given below in decode_syscall
have a simple transformation using tracer_copy_{in,out}.
Read, write, and similar syscalls need better performance
and thus have more specialized implementations shown here.
*/

/*
SYSCALL_read and friends are implemented by loading the data
into the channel, and then redirecting the system call
to read from the channel fd. SYSCALL_readv is done exactly
the same as read, but only the first chunk is actually
read.  The caller must examine the result and then keep reading.
*/

static void decode_read( struct pfs_process *p, int entering, int syscall, int *args )
{
	int fd = args[0];
	void *uaddr = (void*) args[1];
	pfs_size_t length = args[2];
	pfs_off_t offset = args[3];
	char *local_addr;
	
	if(entering) {
		if(!pfs_channel_alloc(length,&p->io_channel_offset)) {
			divert_to_dummy(p,-ENOMEM);
			return;
		}
		local_addr = pfs_channel_base() + p->io_channel_offset;

		if(syscall==SYSCALL_read) {
			p->syscall_result = pfs_read(fd,local_addr,length);
		} else if(syscall==SYSCALL_pread) {
			p->syscall_result = pfs_pread(fd,local_addr,length,offset);
		} else if(syscall==SYS_RECV) {
			p->syscall_result = pfs_recv(fd,local_addr,length,args[3]);
		} else if(syscall==SYS_RECVFROM) {
			p->syscall_result = pfs_recvfrom(fd,local_addr,length,args[3],(struct sockaddr *)args[4],(int*)args[5]);
		}

		p->diverted_length = 0;

		if(p->syscall_result==0) {
			divert_to_dummy(p,0);
		} else if(p->syscall_result>0) {
			divert_to_channel(p,SYSCALL_pread,uaddr,p->syscall_result,p->io_channel_offset);
			pfs_read_count += p->syscall_result;
		} else if( errno==EAGAIN ) {
			if(p->interrupted) {
				p->interrupted = 0;
				divert_to_dummy(p,-EINTR);
			} else if(pfs_is_nonblocking(fd)) {
				divert_to_dummy(p,-EAGAIN);
			} else {
				pfs_channel_free(p->io_channel_offset);
				p->state = PFS_PROCESS_STATE_WAITREAD;
				int rfd = pfs_get_real_fd(fd);
				if(rfd>=0) pfs_poll_wakeon(rfd,PFS_POLL_READ);
			}
		} else {
			divert_to_dummy(p,-errno);
		}
	} else {
		/*
		This is an ugly situation.
		If we arrive here with EINTR, it means that we have copied
		all of the data into the channel, taken any side effects
		of accessing the remote storage device, but the system call
		happened not to read it because of an incoming signal.
		We have no way of re-trying the channel read, so we do
		the ugly slow copy out instead.
		*/

		if( (p->syscall_result==-EINTR) && (p->diverted_length>0) ) {
			tracer_copy_out(p->tracer,pfs_channel_base()+p->io_channel_offset,uaddr,p->diverted_length);
			p->syscall_result = p->diverted_length;
			tracer_result_set(p->tracer,p->syscall_result);
		}

		pfs_channel_free(p->io_channel_offset);
	}
}

/*
decode_write is much the same as read.  We allocate space
in the channel, and then redirect the caller to write
to it.  When the syscall completes, we write the data
to its destination and then set the result.
*/

static void decode_write( struct pfs_process *p, int entering, int syscall, int *args )
{
	if(entering) {
		void *uaddr = (void*) args[1];
		int length = args[2];
		if(!pfs_channel_alloc(length,&p->io_channel_offset)) {
			divert_to_dummy(p,-ENOMEM);
			return;
		}
		divert_to_channel(p,SYSCALL_pwrite,uaddr,length,p->io_channel_offset);
	} else {
		int actual_result;
		tracer_result_get(p->tracer,&actual_result);

		if(actual_result>0) {
			int fd = args[0];
			pfs_off_t offset = args[3];
			char *local_addr = pfs_channel_base() + p->io_channel_offset;

			if(syscall==SYSCALL_write) {
				p->syscall_result = pfs_write(fd,local_addr,actual_result);
			} else if(syscall==SYSCALL_pwrite) {
				p->syscall_result = pfs_pwrite(fd,local_addr,actual_result,offset);
			} else if(syscall==SYS_SEND) {
				p->syscall_result = pfs_send(fd,local_addr,actual_result,args[3]);
			} else if(syscall==SYS_SENDTO) {
				p->syscall_result = pfs_sendto(fd,local_addr,actual_result,args[3],(struct sockaddr *)args[4],args[5]);
			}

			if(p->syscall_result>=0) {
				tracer_result_set(p->tracer,p->syscall_result);
				pfs_channel_free(p->io_channel_offset);
				p->state = PFS_PROCESS_STATE_KERNEL;
				entering = 0;
				pfs_write_count += p->syscall_result;
			} else {
				if(errno==EAGAIN && !pfs_is_nonblocking(fd)) {
					p->state = PFS_PROCESS_STATE_WAITWRITE;
					int rfd = pfs_get_real_fd(fd);
					if(rfd>=0) pfs_poll_wakeon(rfd,PFS_POLL_WRITE);
				} else {
					p->syscall_result = -errno;
					tracer_result_set(p->tracer,p->syscall_result);
					pfs_channel_free(p->io_channel_offset);
					if(p->syscall_result==-EPIPE) {
						// make sure that we are not in a wait state,
						// otherwise pfs_process_raise will re-dispatch.
						p->state = PFS_PROCESS_STATE_KERNEL;
						pfs_process_raise(p->pid,SIGPIPE,1);
					}
				}
			}
		} else {
			pfs_channel_free(p->io_channel_offset);
		}
	}
}

static struct iovec * iovec_alloc_in( struct pfs_process *p, struct iovec *uv, int count )
{
	struct iovec *v;
	int size = sizeof(struct iovec)*count;

	v = (struct iovec *) malloc(size);
	if(v) {
		tracer_copy_in(p->tracer,v,uv,size);
		return v;
	} else {
		return 0;
	}
}

static int iovec_size( struct pfs_process *p, struct iovec *v, int count )
{
	int i, total=0;
	for(i=0;i<count;i++) {
		total += v[i].iov_len;
	}
	return total;	
}

static int iovec_copy_in( struct pfs_process *p, char *buf, struct iovec *v, int count )
{
	int i, pos=0;
	for(i=0;i<count;i++) {
		tracer_copy_in(p->tracer,&buf[pos],(void*)v[i].iov_base,v[i].iov_len);
		pos += v[i].iov_len;
	}
	return pos;
}

static int iovec_copy_out( struct pfs_process *p, char *buf, struct iovec *v, int count )
{
	int i, pos=0;
	for(i=0;i<count;i++) {
		tracer_copy_out(p->tracer,&buf[pos],(void*)v[i].iov_base,v[i].iov_len);
		pos += v[i].iov_len;
	}
	return pos;
}

/*
Both readv and writev have a careful but inefficient implementation.
For each uio block, we examine the data structures, do a manual
read and write in our local buffer, and then copy the data over.
I assume that these are not heavily used system calls, although
they do seem to appear sporadically in X11, the dynamic linker,
and sporadically in networking utilities.
*/

static void decode_readv( struct pfs_process *p, int entering, int syscall, int *args )
{
	if(entering) {
		int fd = args[0];
		struct iovec *uv = (struct iovec *) args[1];
		int count = args[2];

		struct iovec *v;
		int size;
		char *buffer;
		int total = 0;
		int chunk;
		
		if(!uv || count<=0) {
			divert_to_dummy(p,-EINVAL);
			return;
		}

		v = iovec_alloc_in(p,uv,count);
		if(v) {
			size = iovec_size(p,v,count);
			buffer = (char*) malloc(size);
			if(buffer) {
				while(size>0) {
					chunk = pfs_read(fd,&buffer[total],size);
					if(chunk>0) {
						total += chunk;
						size -= chunk;
					} else if(chunk==0) {
						break;
					} else if(chunk<0) {
						if(errno==EAGAIN && !pfs_is_nonblocking(fd)) {
							continue;
						} else {
							break;
						}
					}		
				}

				if(total>0) {
					iovec_copy_out(p,buffer,v,count);
					divert_to_dummy(p,total);
				} else if(chunk==0) {
					divert_to_dummy(p,0);
				} else {
					divert_to_dummy(p,-errno);
				}
				free(buffer);
			} else {
				divert_to_dummy(p,-ENOMEM);
			}
			free(v);		
		} else {
			divert_to_dummy(p,-ENOMEM);
		}
	}
}

static void decode_writev( struct pfs_process *p, int entering, int syscall, int *args )
{
	if(entering) {
		int fd = args[0];
		struct iovec *uv = (struct iovec *) args[1];
		int count = args[2];

		struct iovec *v;
		int size;
		char *buffer;
		int total = 0;
		int chunk;

		if(!uv || count<=0) {
			divert_to_dummy(p,-EINVAL);
			return;
		}

		v = iovec_alloc_in(p,uv,count);
		if(v) {
			size = iovec_size(p,v,count);
			buffer = (char *) malloc(size);
			if(buffer) {
				iovec_copy_in(p,buffer,v,count);
				while(size>0) {
					chunk = pfs_write(fd,&buffer[total],size);
					if(chunk>0) {
						total += chunk;
						size -= chunk;
					} else if(chunk==0) {
						break;
					} else if(chunk<0) {
						if(errno==EAGAIN && !pfs_is_nonblocking(fd)) {
							continue;
						} else {
							break;
						}
					}		
				}

				if(total>0) {
					divert_to_dummy(p,total);
				} else if(chunk==0) {
					divert_to_dummy(p,0);
				} else {
					divert_to_dummy(p,-errno);
				}
				free(buffer);
			} else {
				divert_to_dummy(p,-ENOMEM);
			}
			free(v);		
		} else {
			divert_to_dummy(p,-ENOMEM);
		}
	}
}

static void decode_stat( struct pfs_process *p, int entering, int syscall, int *args, int sixty_four )
{
	int fd = args[0];
	void *uaddr = (void*) args[1];
	char *local_addr;
	struct pfs_stat lbuf;
	struct pfs_kernel_stat kbuf;
	struct pfs_kernel_stat64 kbuf64;
	char path[PFS_PATH_MAX];
	int bufsize;

	if(entering) {
		p->io_channel_offset = 0;

		if(syscall==SYSCALL_stat) {
			tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
			p->syscall_result = pfs_stat(path,&lbuf);
		} else if(syscall==SYSCALL_lstat) {
			tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
			p->syscall_result = pfs_lstat(path,&lbuf);
		} else if(syscall==SYSCALL_fstat) {
			p->syscall_result = pfs_fstat(fd,&lbuf);
		}

		if(p->syscall_result>=0) {
			if(!pfs_channel_alloc(sizeof(kbuf64),&p->io_channel_offset)) {
				divert_to_dummy(p,-ENOMEM);
			} else {
				local_addr = pfs_channel_base() + p->io_channel_offset;

				if(sixty_four) {
					COPY_STAT(lbuf,kbuf64);
					/* Special case: Linux needs stat64.st_ino in two places. */
					kbuf64.st_ino_extra = kbuf64.st_ino;
					memcpy(local_addr,&kbuf64,sizeof(kbuf64));
					bufsize = sizeof(kbuf64);
				} else {
					COPY_STAT(lbuf,kbuf);
					memcpy(local_addr,&kbuf,sizeof(kbuf));
					bufsize = sizeof(kbuf);
				}
				divert_to_channel(p,SYSCALL_pread,uaddr,bufsize,p->io_channel_offset);
			}
		} else {
			divert_to_dummy(p,-errno);
		}
	} else {
		if(p->io_channel_offset!=0) {
			pfs_channel_free(p->io_channel_offset);
			divert_to_dummy(p,0);
		}
	}
}

static void decode_statfs( struct pfs_process *p, int entering, int syscall, int *args )
{
	int fd = args[0];
	void *uaddr = (void*) args[1];
	char *local_addr;
	struct pfs_statfs lbuf;
	struct pfs_kernel_statfs kbuf;
	char path[PFS_PATH_MAX];
	int bufsize;

	if(entering) {
		p->io_channel_offset = 0;

		if(syscall==SYSCALL_statfs) {
			tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
			p->syscall_result = pfs_statfs(path,&lbuf);
		} else if(syscall==SYSCALL_fstatfs) {
			p->syscall_result = pfs_fstatfs(fd,&lbuf);
		}

		if(p->syscall_result>=0) {
			if(!pfs_channel_alloc(sizeof(kbuf),&p->io_channel_offset)) {
				divert_to_dummy(p,-ENOMEM);
			} else {
				local_addr = pfs_channel_base() + p->io_channel_offset;
				COPY_STATFS(lbuf,kbuf);
				memcpy(local_addr,&kbuf,sizeof(kbuf));
				bufsize = sizeof(kbuf);
				divert_to_channel(p,SYSCALL_pread,uaddr,bufsize,p->io_channel_offset);
			}
		} else {
			divert_to_dummy(p,-errno);
		}
	} else {
		if(p->io_channel_offset!=0) {
			pfs_channel_free(p->io_channel_offset);
			divert_to_dummy(p,0);
		}
	}
}

/*
On Linux, all of the socket related system calls are
multiplexed through one system call.
*/

void decode_socketcall( struct pfs_process *p, int entering, int syscall, int *a )
{
	int length;
	int fds[2];
	void *x;
	int r;
	int t4;
	int t5;

	struct sockaddr_un *paddr;
	struct sockaddr_un addr;
 
	if(entering) {
		switch(syscall) {
		case SYS_SOCKET:
			r = pfs_socket(a[0],a[1],a[2]);
			if(r<0) r=-errno;
			divert_to_dummy(p,r);
			break;
		case SYS_BIND:
			x = xxmalloc(a[2]+2);
			tracer_copy_in(p->tracer,x,(void*)a[1],a[2]);
			paddr = (struct sockaddr_un *)x;
			if(paddr->sun_family==AF_UNIX) {
				pfs_name pname;
				((char*)x)[a[2]] = 0;
				if(pfs_resolve_name(paddr->sun_path,&pname)) {
					addr.sun_family = AF_UNIX;
					strcpy(addr.sun_path,pname.path);
					r = pfs_bind(a[0],(struct sockaddr *)&addr,sizeof(addr));
				} else {
					r = -1;
				}
			} else {
				r = pfs_bind(a[0],(struct sockaddr *)x,a[2]);
			}
			if(r<0) r=-errno;
			free(x);
			divert_to_dummy(p,r);
			break;
		case SYS_CONNECT:
			x = xxmalloc(a[2]+2);
			tracer_copy_in(p->tracer,x,(void*)a[1],a[2]);
			paddr = (struct sockaddr_un *)x;
			if(paddr->sun_family==AF_UNIX) {
				pfs_name pname;
				((char*)x)[a[2]] = 0;
				if(pfs_resolve_name(paddr->sun_path,&pname)) {
					addr.sun_family = AF_UNIX;
					strcpy(addr.sun_path,pname.path);
					r = pfs_connect(a[0],(struct sockaddr *)&addr,sizeof(addr));
				} else {
					r = -1;
				}
			} else {
				r = pfs_connect(a[0],(struct sockaddr*)x,a[2]);
			}
			if(r>=0) {
				divert_to_dummy(p,r);
			} else if(errno_in_progress(errno)) {
				if(p->interrupted) {
					p->interrupted = 0;
					divert_to_dummy(p,-EINTR);
				} else if(pfs_is_nonblocking(a[0])) {
					divert_to_dummy(p,-EINPROGRESS);
				} else {
					p->state = PFS_PROCESS_STATE_WAITREAD;
					int rfd = pfs_get_real_fd(a[0]);
					if(rfd>=0) pfs_poll_wakeon(rfd,PFS_POLL_READ|PFS_POLL_WRITE|PFS_POLL_EXCEPT);
				}
			} else {
				divert_to_dummy(p,-errno);
			}
			free(x);
			break;
		case SYS_LISTEN:
			r = pfs_listen(a[0],a[1]);
			if(r<0) r=-errno;
			divert_to_dummy(p,r);
			break;
		case SYS_ACCEPT:
			if(a[1]) {
				tracer_copy_in(p->tracer,&length,(void*)a[2],sizeof(length));
				x = xxmalloc(length);
				r = pfs_accept(a[0],(struct sockaddr*)x,&length);
			} else {
				x = 0;
				r = pfs_accept(a[0],0,0);
			}
			if(r>=0) {
				if(x) {
					tracer_copy_out(p->tracer,x,(void*)a[1],length);
					tracer_copy_out(p->tracer,&length,(void*)a[2],sizeof(length));
				}
				divert_to_dummy(p,r);
			} else if(errno_in_progress(errno)) {
				if(p->interrupted) {
					p->interrupted = 0;
					divert_to_dummy(p,-EINTR);
				} else if(pfs_is_nonblocking(a[0])) {
					divert_to_dummy(p,-EAGAIN);
				} else {
					p->state = PFS_PROCESS_STATE_WAITREAD;
					int rfd = pfs_get_real_fd(a[0]);
					if(rfd>=0) pfs_poll_wakeon(rfd,PFS_POLL_READ);
				}
			} else {
				divert_to_dummy(p,-errno);
			}
			if(x) free(x);
			break;
		case SYS_GETSOCKNAME:
			tracer_copy_in(p->tracer,&length,(void*)a[2],sizeof(length));
			x = xxmalloc(length);
			r = pfs_getsockname(a[0],(struct sockaddr *)x,&length);
			if(r<0) {
				r=-errno;
			} else {
				tracer_copy_out(p->tracer,x,(void*)a[1],length);
				tracer_copy_out(p->tracer,&length,(void*)a[2],sizeof(length));
			}
			free(x);
			divert_to_dummy(p,r);
			break;
		case SYS_GETPEERNAME:
			tracer_copy_in(p->tracer,&length,(void*)a[2],sizeof(length));
			x = xxmalloc(length);
			r = pfs_getpeername(a[0],(struct sockaddr *)x,&length);
			if(r<0) {
				r=-errno;
			} else {
				tracer_copy_out(p->tracer,x,(void*)a[1],length);
				tracer_copy_out(p->tracer,&length,(void*)a[2],sizeof(length));
			}
			free(x);
			divert_to_dummy(p,r);
			break;
		case SYS_SOCKETPAIR:
			r = pfs_socketpair(a[0],a[1],a[2],fds);
			if(r<0) {
				r=-errno;
			} else {
				tracer_copy_out(p->tracer,fds,(void*)a[3],sizeof(fds));
			}
			divert_to_dummy(p,r);
			break;
		case SYS_SEND:
			decode_write(p,entering,syscall,a);
			break;
		case SYS_SENDTO:
			if(a[4]) {
				x = xxmalloc(a[5]);
				tracer_copy_in(p->tracer,x,(void*)a[4],a[5]);
				t4 = a[4];
				a[4] = (int)x;
			}
			decode_write(p,entering,syscall,a);
			if(a[4]) {
				a[4] = t4;
				free(x);
			}
			break;
		case SYS_RECV:
			decode_read(p,entering,syscall,a);
			break;
		case SYS_RECVFROM:
			if(a[4]) {
				tracer_copy_in(p->tracer,&length,(void*)a[5],sizeof(length));
				x = xxmalloc(length);
				t4 = a[4];
				t5 = a[5];
				a[4] = (int) x;
				a[5] = (int) &length;
			}
			decode_read(p,entering,syscall,a);
			if(a[4]) {
				a[4] = t4;
				a[5] = t5;
				if(p->syscall_result>=0) {
					tracer_copy_out(p->tracer,x,(void*)a[4],length);
					tracer_copy_out(p->tracer,&length,(void*)a[5],sizeof(length));
				}
				free(x);
			}
			break;
		case SYS_SHUTDOWN:
			r = pfs_shutdown(a[0],a[1]);
			if(r<0) r=-errno;
			divert_to_dummy(p,r);
			break;
		case SYS_SETSOCKOPT:
			x = xxmalloc(a[4]);
			tracer_copy_in(p->tracer,x,(void*)a[3],a[4]);
			r = pfs_setsockopt(a[0],a[1],a[2],x,a[4]);
			if(r<0) r=-errno;
			free(x);
			divert_to_dummy(p,r);
			break;
		case SYS_GETSOCKOPT:
			tracer_copy_in(p->tracer,&length,(void*)a[4],sizeof(length));
			x = xxmalloc(length);
			r = pfs_getsockopt(a[0],a[1],a[2],x,&length);
			if(r<0) {
				r=-errno;
			} else {
				tracer_copy_out(p->tracer,x,(void*)a[3],length);
				tracer_copy_out(p->tracer,&length,(void*)a[4],sizeof(length));
			}		
			free(x);
			divert_to_dummy(p,r);
			break;

		/*
		Holy cow, these are complicated.
		Let's hold off on implementing them until they are needed.
		*/

		case SYS_SENDMSG:
			debug(D_SYSCALL|D_NOTICE,"system call SYS_sendmsg not supported");
			divert_to_dummy(p,-ENOSYS);
			break;
		case SYS_RECVMSG:
			debug(D_SYSCALL|D_NOTICE,"system call SYS_recvmsg not supported");
			divert_to_dummy(p,-ENOSYS);
			break;
		default:
			r = -EINVAL;
			divert_to_dummy(p,r);
			break;
		}
	} else {
		/* Only these few have any action when exiting */

		switch(syscall) {
			case SYS_SEND:
			case SYS_SENDTO:
				decode_write(p,entering,syscall,a);
				break;
			case SYS_RECV:
			case SYS_RECVFROM:
				decode_read(p,entering,syscall,a);
				break;
		}
	}
}
                                                                 
/*
This function is an inexpensive test to see if a given
filename is executable.  It is not all-inclusive, nor should
it be considered a reliable security device.  This function is
over-optimistic in some cases, but if it falsely reports
true, the later real execve() may still fail.
*/

static int is_executable( const char *path)
{
	struct pfs_stat buf;

	if(pfs_stat(path,&buf)!=0) return 0;

	if(buf.st_mode&S_ISUID || buf.st_mode&S_ISGID) {
		debug(D_NOTICE,"cannot execute the program %s because it is setuid.",path);
		errno = EACCES;
		return 0;
	}

	if(buf.st_mode&S_IXUSR || buf.st_mode&S_IXGRP || buf.st_mode&S_IXOTH) {
		return 1;
	} else {
		errno = ENOEXEC;
		return 0;
	}
}


/*
Several things to note about exec.

An entry to execve looks like a normal syscall.
An exit from execve indicates a successfull execve in progress.
Finally, a *third* event with args[0]==0 indicates an execve
that has completed with the new image active.

Now, we cannot execute the path named by the execve directly.
It must be resolved through PFS, because our idea of the 
current dir (or even the meaning of the name) may be quite
different.  We resolve the file name into a local path,
perhaps by pulling it into the cache.

In the simple (second) case, we copy the new local name
into the address space of the process and exec that instead.
If the exec fails, we must restore the changed bytes afterwards.

In the complex (first) case, the program contains a pound-bang
indicating an interpreter.  We instead resolve the interpreter
as the executable and fiddle around with the job's argv to
indicate that.  Then, we do much the same as the first case.
*/

void decode_execve( struct pfs_process *p, int entering, int syscall, int *args ) {
	if(args[0]==0) {
		debug(D_PROCESS,"execve: %s executing ",p->name);
		p->state = PFS_PROCESS_STATE_USER;
	} else if(entering) {
		char path[PFS_PATH_MAX];
		char firstline[PFS_PATH_MAX];

		tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));

		p->new_logical_name[0] = 0;
		p->new_physical_name[0] = 0;

		if(!is_executable(path)) {
			divert_to_dummy(p,-errno);
			return;
		}

		firstline[0] = 0;
		strcpy(p->new_logical_name,path);
		if(pfs_get_local_name(path,p->new_physical_name,firstline,sizeof(firstline))<0) {
			divert_to_dummy(p,-errno);
			return;
		}

		/* remove any newlines or spaces at the end */

		char *c = firstline;
		while(*c) c++;
		c--;
		while( *c=='\n' || *c==' ' ) {
			*c = 0;
			c--;
		}

		if(!strncmp(firstline,"#!",2)) {
			char *argv[PFS_ARG_MAX], *ext_argv;
			char *interp, *ext_interp;
			char *interparg, *ext_interparg;
			char *scriptarg, *ext_scriptarg;
			char *ext_physical_name;
			int i, argc, shiftargs;

			debug(D_PROCESS,"execve: %s is an interpreted executable",p->new_logical_name);

			/* interp points to the interpreter */
			/* store it in the scratch area */

			interp = &firstline[2];
			while(isspace(*interp)) interp++;
			ext_interp = PFS_SCRATCH_ADDR;
			
			/* interparg points to the internal argument */
			/* scriptarg points to the script itself */
			interparg = strchr(interp,' ');
			if(interparg) {
				*interparg = 0;
				interparg++;	
				while(isspace(*interparg)) interparg++;
				ext_interparg = ext_interp + strlen(interp) + 1;
				scriptarg = path;
				ext_scriptarg = ext_interparg + strlen(interparg) + 1;
				debug(D_PROCESS,"execve: instead do %s %s %s",interp,interparg,scriptarg);
				shiftargs = 2;
			} else {
				interparg = 0;
				ext_interparg = ext_interp;
				scriptarg = path;
				ext_scriptarg = ext_interparg;
				shiftargs = 1;
				debug(D_PROCESS,"execve: instead do %s %s",interp,scriptarg);
			}


			/* make sure the new interp is loaded */
			strcpy(p->new_logical_name,interp);
			if(pfs_get_local_name(interp,p->new_physical_name,0,0)!=0) {
				divert_to_dummy(p,-errno);
				return;
			}

			/* the physical name of the interp is next */
			ext_physical_name = ext_scriptarg + strlen(scriptarg) + 1;

			/* the new argv goes in the scratch area next */
			ext_argv = ext_physical_name + strlen(p->new_physical_name) + 1;

			/* load in the arguments given by the program and count them up */
			tracer_copy_in(p->tracer,argv,(void*)args[1],sizeof(argv));
			for(argc=0;argv[argc] && argc<PFS_ARG_MAX;argc++) {}

			/* save the scratch area */
			tracer_copy_in(p->tracer,p->scratch_data,PFS_SCRATCH_ADDR,PFS_SCRATCH_SIZE);

			/* write out the new interp, arg, and physical name */
			tracer_copy_out(p->tracer,interp,ext_interp,strlen(interp)+1);
			if(interparg) tracer_copy_out(p->tracer,interparg,ext_interparg,strlen(interparg)+1);
			tracer_copy_out(p->tracer,scriptarg,ext_scriptarg,strlen(scriptarg)+1);
			tracer_copy_out(p->tracer,p->new_physical_name,ext_physical_name,strlen(p->new_physical_name)+1);
			/* rebuild the argv copy it out */
			for(i=argc-1+shiftargs;i>0;i--) argv[i] = argv[i-shiftargs];
			argc+=shiftargs;
			argv[0] = ext_interp;
			if(interparg) {
				argv[1] = ext_interparg;
				argv[2] = ext_scriptarg;
			} else {
				argv[1] = ext_scriptarg;
			}
			argv[argc] = 0;
			for(i=0;i<=argc;i++) {
				tracer_copy_out(p->tracer,&argv[i],ext_argv+sizeof(char*)*i,sizeof(char*));
			}

			/* change the registers to reflect argv */
			args[0] = (int) ext_physical_name;
			args[1] = (int) ext_argv;
			tracer_args_set(p->tracer,p->syscall,args,3);
		} else {
			debug(D_PROCESS,"execve: %s is an ordinary executable",p->new_logical_name);

			/* save all of the data we are going to clobber */
			tracer_copy_in(p->tracer,p->scratch_data,PFS_SCRATCH_ADDR,PFS_SCRATCH_SIZE);

			/* store the new local path */
			tracer_copy_out(p->tracer,p->new_physical_name,PFS_SCRATCH_ADDR,strlen(p->new_physical_name)+1);

			/* set the new program name to the logical name */
			args[0] = (int)PFS_SCRATCH_ADDR;
			tracer_args_set(p->tracer,p->syscall,args,3);
		}
		debug(D_PROCESS,"execve: %s attempting",p->new_logical_name);
	} else {
		int actual_result;
		tracer_result_get(p->tracer,&actual_result);
		if(actual_result==0) {
			debug(D_PROCESS,"execve: %s working",p->new_logical_name);
			strcpy(p->name,p->new_logical_name);
			/* after a successful exec, signal handlers are reset */
			memset(p->signal_interruptible,0,sizeof(p->signal_interruptible));
			/* and certain files in the file table are closed */
			p->table->close_on_exec();
			debug(D_PSTREE,"%d exec %s",p->pid,p->new_logical_name);
		} else if(p->new_logical_name[0]) {
			debug(D_PROCESS,"execve: %s failed",p->new_logical_name);
			debug(D_PROCESS,"execve: restoring scratch area");
			tracer_copy_out(p->tracer,p->scratch_data,(void*)PFS_SCRATCH_ADDR,PFS_SCRATCH_SIZE);
		}
	}
}


/*
Memory mapped files are loaded into the channel,
the whole file regardless of what portion is actually
mapped.  The channel cache keeps a reference count.
*/

void decode_mmap( struct pfs_process *p, int syscall, int entering, int *args )
{
	if(entering) {
		int fd, flags;
		pfs_size_t length, channel_offset, source_offset;
		int nargs[TRACER_ARGS_MAX];

#ifdef PFS_INDIRECT_MMAP_ARGS
		if(p->syscall==SYSCALL_mmap) {
			tracer_copy_in(p->tracer,nargs,(void*)args[0],sizeof(nargs));
		} else
#endif
		{
			memcpy(nargs,p->syscall_args,sizeof(p->syscall_args));
		}
		flags = nargs[3];
		fd = nargs[4];
		source_offset = nargs[5];

		debug(D_SYSCALL,"mmap flags=%d fd=%d source_offset=%d",flags,fd,source_offset);
		if(flags&MAP_ANONYMOUS) {
			/* great, just do it. */
			debug(D_SYSCALL,"mmap skipped b/c anonymous");
		} else {
			char file_name[PFS_PATH_MAX];

			if(pfs_get_full_name(fd,file_name)!=0) {
				debug(D_SYSCALL,"mmap failed name: %s",strerror(errno));
				divert_to_dummy(p,-errno);
				return;
			}

			if(pfs_channel_cache_alloc(file_name,fd,&length,&channel_offset)) {
				nargs[4] = pfs_channel_fd();
				nargs[5] = channel_offset+source_offset;

#ifdef PFS_INDIRECT_MMAP_ARGS
				if(p->syscall==SYSCALL_mmap) {
					tracer_copy_out(p->tracer,nargs,(void*)args[0],sizeof(nargs));
				} else 
#endif
				{
#ifdef PFS_MMAP_OFFSET_IN_PAGES
					nargs[5] = nargs[5] / getpagesize();
#endif
					tracer_args_set(p->tracer,p->syscall,nargs,6);
					p->syscall_args_changed = 1;
				}	
				debug(D_SYSCALL,"mmap changed to %d %d",nargs[4],nargs[5]);
			} else {
				debug(D_SYSCALL,"mmap failed cache: %s",strerror(errno));
				divert_to_dummy(p,-errno);
				return;
			}
		}
	}
}

int decode_ioctl_siocgifconf( struct pfs_process *p, int fd, int cmd, void *uaddr )
{
	struct ifconf uifc;
	struct ifconf ifc;
	char *buffer;
	int length;
	int result;

	tracer_copy_in(p->tracer,&uifc,uaddr,sizeof(uifc));
	buffer = (char*) malloc(uifc.ifc_len);
	tracer_copy_in(p->tracer,buffer,uifc.ifc_buf,uifc.ifc_len);

	ifc.ifc_buf = buffer;
	ifc.ifc_len = length;

	result = pfs_ioctl(fd,cmd,&ifc);

	if(result>=0) {
		uifc.ifc_len = ifc.ifc_len;
		tracer_copy_out(p->tracer,&uifc,uaddr,sizeof(uifc));
		tracer_copy_out(p->tracer,buffer,uifc.ifc_buf,uifc.ifc_len);
	}

	free(buffer);
	
	return result;
}

void decode_syscall( struct pfs_process *p, int entering )
{
	int *args;

	char path[PFS_PATH_MAX];
	char path2[PFS_PATH_MAX];

	if(entering) {
		p->state = PFS_PROCESS_STATE_KERNEL;
		p->syscall_dummy = 0;
		tracer_args_get(p->tracer,&p->syscall,p->syscall_args);
		debug(D_SYSCALL,"%s",tracer_syscall_name(p->tracer,p->syscall));
		p->syscall_original = p->syscall;
		pfs_syscall_count++;
	}

	args = p->syscall_args;

	switch(p->syscall) {
		case SYSCALL_execve:
			decode_execve(p,entering,p->syscall,args);
			break;

		/*
		Some variants of fork do not propagate ptrace, so we
		must convert them into clone with appropriate flags.
		Once a fork is started, we must trace only that pid
		so that we can determine the child pid before seeing
		any events from the child. On return, we must fill
		in the child process with its parent's ppid.
		*/

		case SYSCALL_fork:
		case SYSCALL_clone:
			if(entering) {
				int newargs[4];
				int newargs_count;
				if(p->syscall==SYSCALL_fork) {
					newargs[0] = CLONE_PTRACE|CLONE_PARENT|SIGCHLD;
					newargs[1] = 0;
					newargs_count = 2;
					p->syscall_args_changed = 1;
					debug(D_SYSCALL,"converting fork into clone(%x)",newargs[0]);
				} else {
					newargs[0] = (args[0]&~0xff)|CLONE_PTRACE|CLONE_PARENT|SIGCHLD;
					newargs_count = 1;
					debug(D_SYSCALL,"adjusting clone(%x,%x,%x,%x) -> clone(%x)",args[0],args[1],args[2],args[3],newargs[0]);
				}
				tracer_args_set(p->tracer,SYSCALL_clone,newargs,newargs_count);
				trace_this_pid = p->pid;
			} else {
				pid_t childpid;
				struct pfs_process *child;
				tracer_result_get(p->tracer,&childpid);
				if(childpid>0) {
					int child_signal,clone_files;
					if(p->syscall_original==SYSCALL_fork) {
						child_signal = SIGCHLD;
						clone_files = 0;
					} else {
						child_signal = args[0]&0xff;
						clone_files = args[0]&CLONE_FILES;
					}
					child = pfs_process_create(childpid,p->pid,clone_files,child_signal);
					child->syscall_result = 0;
					if(p->syscall_original==SYSCALL_fork) {
						memcpy(child->syscall_args,p->syscall_args,sizeof(p->syscall_args));
						child->syscall_args_changed = 1;
					}
					if(pfs_trap_after_fork) {
						child->state = PFS_PROCESS_STATE_KERNEL;
					} else {
						child->state = PFS_PROCESS_STATE_USER;
					}
					debug(D_PROCESS,"%d created pid %d",p->pid,childpid);
					/* now trace any process at all */
					trace_this_pid = -1;
				}

			}
			break;

		/*
		Note that we do not support vfork.  The behavior of vfork
		varies greatly from kernel to kernel, and is in fact impossible
		to support through ptrace without a kernel patch in some cases.
		However, glibc is smart and converts vfork into fork if the
		kernel response that it does not exist.  So, failed vforks
		eventually end up in the previous case.  Also note parrot_helper.so,
		which also aims to solve this problem.
		*/

		case SYSCALL_vfork:
			if(entering) {
				divert_to_dummy(p,-ENOSYS);
			}
			break;

		/*
		On ptraces that do not preserve the process structure,
		we must trap and manage variants of wait() to permit
		the propagation of child completion.
		*/
#ifdef PFS_MANAGE_PROCESS_TREES
		case SYSCALL_waitpid:
			if(entering) {
				pfs_process_waitpid(p,args[0],(int*)args[1],args[2],0);
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_wait4:
			if(entering) {
				pfs_process_waitpid(p,args[0],(int*)args[1],args[2],(struct rusage*)args[3]);
				divert_to_dummy(p,p->syscall_result);
			}
			break;
#endif

		/*
		We don't do anything special with exit.  Just let it
		run to completion, and then process the exit event
		in the main loop.
		*/

		case SYSCALL_exit:
		case SYSCALL_exit_group:
			break;

		/*
		Here begin all of the I/O operations, given in the
		same order as in pfs_table.  Notice that most operations
		use the simple but slow tracer_copy_{in,out} routines.
		When performance is important (write,mmap), we resort
		to redirection I/O to the side channel.
		*/

		/* File descriptor creation */

		case SYSCALL_open:
		case SYSCALL_creat:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				if(p->syscall==SYSCALL_creat) {
					p->syscall_result = pfs_open(path,O_CREAT|O_WRONLY|O_TRUNC,args[1]);
				} else {
					p->syscall_result = pfs_open(path,args[1],args[2]);
				}
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_pipe:
			if(entering) {
				int fds[2];
				p->syscall_result = pfs_pipe(fds);
				if(p->syscall_result<0) {
					p->syscall_result = -errno;
				} else {
					tracer_copy_out(p->tracer,(void*)fds,(void*)args[0],sizeof(fds));
				}
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		/* operations on open files */

		case SYSCALL_close:
			if(entering) {
				p->syscall_result = pfs_close(args[0]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_read:
		case SYSCALL_pread:
			decode_read(p,entering,p->syscall,args);
			break;

		case SYSCALL_write:
		case SYSCALL_pwrite:
			decode_write(p,entering,p->syscall,args);
			break;

		case SYSCALL_readv:
			decode_readv(p,entering,p->syscall,args);
			break;

		case SYSCALL_writev:
			decode_writev(p,entering,p->syscall,args);
			break;

		#ifdef SYSCALL_socketcall
		case SYSCALL_socketcall: {
			int subargs[6];
			tracer_copy_in(p->tracer,subargs,(void*)args[1],sizeof(subargs));
			decode_socketcall(p,entering,args[0],subargs);
			break;
		}
		#endif

		/*
		Now we have a series of standard file operations that
		only use the integer arguments, and are (mostly) easily
		passed back and forth.
		*/

		case SYSCALL_lseek:
			if(entering) {
				p->syscall_result = pfs_lseek(args[0],args[1],args[2]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		#ifdef SYSCALL__llseek
		case SYSCALL__llseek:
			if(entering) {
				unsigned high = args[1];
				unsigned low = args[2];
				UINT64_T offset = ( (UINT64_T) high << 32 ) | (UINT64_T) low;
				INT64_T result = pfs_lseek(args[0],offset,args[4]);
				if(result<0) {
					p->syscall_result = -errno;
				} else {
					tracer_copy_out(p->tracer,&result,(void*)args[3],sizeof(result));
					p->syscall_result = 0;
				}
				divert_to_dummy(p,p->syscall_result);
			}
			break;
		#endif

		#ifdef SYSCALL_ftruncate
		case SYSCALL_ftruncate:
			if(entering) {
				p->syscall_result = pfs_ftruncate(args[0],args[1]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
		break;
		#endif

		#ifdef SYSCALL_ftruncate64
		case SYSCALL_ftruncate64:
			if(entering) {
				long long offset = args[1]+(((long long)args[2])<<32);
				p->syscall_result = pfs_ftruncate(args[0],offset);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;
		#endif

		case SYSCALL_fstat:
			decode_stat(p,entering,SYSCALL_fstat,args,0);
			break;
		case SYSCALL_fstat64:
			decode_stat(p,entering,SYSCALL_fstat,args,1);
			break;
		case SYSCALL_fstatfs:
			decode_statfs(p,entering,SYSCALL_fstatfs,args);
			break;

		case SYSCALL_flock:
			if(entering) {
				p->syscall_result = pfs_flock(args[0],args[1]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;
				
		case SYSCALL_fsync:
		case SYSCALL_fdatasync:
			if(entering) {
				p->syscall_result = pfs_fsync(args[0]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_fchdir:
			if(entering) {
				p->syscall_result = pfs_fchdir(args[0]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_fchmod:
			if(entering) {
				p->syscall_result = pfs_fchmod(args[0],args[1]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_fchown:
		#ifdef SYSCALL_fchown32
		case SYSCALL_fchown32:
		#endif
			if(entering) {
				p->syscall_result = pfs_fchown(args[0],args[1],args[2]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		/*
		ioctl presents both bad news and good news.
		The bad news is that all ioctl operations are driver
		specific.  I have no intention of coding up all the
		possible ioctls here, nor could I. However, it is a fair
		assumption that the third argument, when a valid pointer,
		is a continuous and small chunk of memory.  So, we copy
		in one page that it points to (if any), and use that as
		a temporary buffer.  We copy back to the application only
		those words in the buffer that change.  If none change,
		then the argument was likely interpreted as an integer
		rather than a pointer.  A nice coincidence is that 
		copy in is generally efficient, while copy out, though
		slow, is minimized.
		*/

		case SYSCALL_ioctl:
			if(entering) {
				int fd = args[0];
				int cmd = args[1];
				void *uaddr = (void*) args[2];
				char buffer[65536];
				char tbuffer[65536];
				int length;


				if(cmd==SIOCGIFCONF) {
					p->syscall_result = decode_ioctl_siocgifconf(p,fd,cmd,uaddr);
					divert_to_dummy(p,p->syscall_result);
					break;
				}

				if(uaddr) {
					length = tracer_copy_in(p->tracer,buffer,uaddr,sizeof(buffer));
					if(length>0) {
						memcpy(tbuffer,buffer,length);
						p->syscall_result = pfs_ioctl(fd,cmd,buffer);
					} else {
						uaddr = 0;
						p->syscall_result = pfs_ioctl(fd,cmd,uaddr);
					}
				} else {
					p->syscall_result = pfs_ioctl(fd,cmd,uaddr);
				}

				if(p->syscall_result<0) {
					p->syscall_result = -errno;
				} else {
					if(uaddr) {
						int i, changed=0;
						for(i=0;i<length;i++) {
							if(tbuffer[i]!=buffer[i]) {
								changed = i+1;
							}
						}
						changed = _ROUND_UP(changed,sizeof(long));
						tracer_copy_out(p->tracer,buffer,uaddr,changed);
					}
				}
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		/*
		Unlike ioctl, fcntl operations are rather generic
		and operate on the file table itself.  These things
		we can parse, understand, and pass along to the 
		file table in most cases.

		We permit the user to set the O_ASYNC flag and thus
		receive activity notification via SIGIO.  However,
		we don't yet support extended signal information.
		*/

		case SYSCALL_fcntl:
		#ifdef SYSCALL_fcntl64
		case SYSCALL_fcntl64:
		#endif
			if(entering) {
				int fd = args[0];
				int cmd = args[1];
				void *uaddr = (void*) args[2];
				struct flock fl;
				struct flock64 fl64;

				switch(cmd) {
					case F_DUPFD:
					case F_GETFD:
					case F_SETFD:
					case F_GETFL:
					case F_SETFL:
						p->syscall_result = pfs_fcntl(fd,cmd,uaddr);
						if(p->syscall_result<0) p->syscall_result=-errno;
						divert_to_dummy(p,p->syscall_result);

						if(cmd==F_SETFL) {
							int flags = (int)args[2];
							if(flags&O_ASYNC) {
								debug(D_PROCESS,"pid %d requests O_ASYNC on fd %d",current->pid,fd);
								p->flags |= PFS_PROCESS_FLAGS_ASYNC;
							}
						}
						break;

					case PFS_GETLK:
					case PFS_SETLK:
					case PFS_SETLKW:
						tracer_copy_in(p->tracer,&fl,uaddr,sizeof(fl));
						p->syscall_result = pfs_fcntl(fd,cmd,&fl);
						if(p->syscall_result<0) {
							p->syscall_result=-errno;
						} else {
							tracer_copy_out(p->tracer,&fl,uaddr,sizeof(fl));
						}
						divert_to_dummy(p,p->syscall_result);
						break;

					case PFS_GETLK64:
					case PFS_SETLK64:
					case PFS_SETLKW64:
						tracer_copy_in(p->tracer,&fl64,uaddr,sizeof(fl64));
						p->syscall_result = pfs_fcntl(fd,cmd,&fl64);
						if(p->syscall_result<0) {
							p->syscall_result=-errno;
						} else {
							tracer_copy_out(p->tracer,&fl64,uaddr,sizeof(fl64));
						}
						divert_to_dummy(p,p->syscall_result);
						break;

					/* Pretend that the caller is the signal recipient */
					case F_GETOWN:
						divert_to_dummy(p,p->pid);
						break;

					/* But we always get the signal. */
					case F_SETOWN:
						debug(D_PROCESS,"pid %d requests F_SETOWN on fd %d",current->pid,fd);
						p->flags |= PFS_PROCESS_FLAGS_ASYNC;
						pfs_fcntl(fd,F_SETOWN,(void*)getpid());
						divert_to_dummy(p,0);
						break;

					default:
						divert_to_dummy(p,-ENOSYS);
						break;
				}
			}
			break;


		case SYSCALL_mmap:
		case SYSCALL_mmap2:
			decode_mmap(p,p->syscall,entering,args);
			break;

		/*
		XXX Incomplete but acceptable for now.
		We permit the application to munmap whatever it likes.
		This might be an anonymous map that we don't track,
		or it could be file.  In the latter case, we would like
		to decrement the reference count.  At the moment, we
		can't tell the difference, nor can we reverse map the
		application's space into our own address space.
		Once loaded, mmaped objects just stay in until the
		parent pfs process dies.
		*/

		case SYSCALL_munmap:
			break;

		/*
		For select, we must copy in all the data structures
		that are pointed to, select, and then copy out.
		Notice that on Linux, newselect has the ordinary
		interface, while SYSCALL_select keeps all of the arguments
		in a single structure.
		*/

		case SYSCALL__newselect:
			if(entering) {
				int maxfd = args[0];
				fd_set rset, wset, eset;
				struct timeval tv;
				fd_set *prset, *pwset, *peset;
				struct timeval *ptv;
				int nlongs,nbytes;

				nlongs = (maxfd+31)/32;
				nbytes = nlongs*4;

				FD_ZERO(&rset);
				FD_ZERO(&wset);
				FD_ZERO(&eset);

				if(args[1]) {
					tracer_copy_in(p->tracer,&rset,(void*)args[1],nbytes);
					prset = &rset;
				} else {
					prset = 0;
				}

				if(args[2]) {
					tracer_copy_in(p->tracer,&wset,(void*)args[2],nbytes);
					pwset = &wset;
				} else {
					pwset = 0;
				}

				if(args[3]) {
					tracer_copy_in(p->tracer,&eset,(void*)args[3],nbytes);
					peset = &eset;
				} else {
					peset = 0;
				}

				if(args[4]) {
					tracer_copy_in(p->tracer,&tv,(void*)args[4],sizeof(tv));
					ptv = &tv;
				} else {
					ptv = 0;
				}

				p->syscall_result = pfs_select(maxfd,prset,pwset,peset,ptv);

				if(p->syscall_result>=0) {
					divert_to_dummy(p,p->syscall_result);
					if(prset) tracer_copy_out(p->tracer,prset,(void*)args[1],nbytes);
					if(pwset) tracer_copy_out(p->tracer,pwset,(void*)args[2],nbytes);
					if(peset) tracer_copy_out(p->tracer,peset,(void*)args[3],nbytes);
					if(ptv) tracer_copy_out(p->tracer,ptv,(void*)args[4],sizeof(*ptv));
				} else if(errno==EAGAIN) {
					if(p->interrupted) {
						p->interrupted = 0;
						divert_to_dummy(p,-EINTR);
					} else {
						p->state = PFS_PROCESS_STATE_WAITREAD;
					}
				} else {
					divert_to_dummy(p,-errno);
				}
			}
			break;

		case SYSCALL_poll:
			if(entering) {
				struct pollfd *ufds;
				if(args[1]>1024) {
					divert_to_dummy(p,-EINVAL);
				} else {
					int length = sizeof(*ufds)*args[1];
					ufds = (struct pollfd *) malloc(length);
					if(ufds) {
						tracer_copy_in(p->tracer,ufds,(void*)args[0],length);
						p->syscall_result = pfs_poll(ufds,args[1],args[2]);
						if(p->syscall_result>=0) {
							divert_to_dummy(p,p->syscall_result);
							tracer_copy_out(p->tracer,ufds,(void*)args[0],length);
						} else if(errno==EAGAIN) {
							if(p->interrupted) {
								p->interrupted = 0;
								divert_to_dummy(p,-EINTR);
							} else {
								p->state = PFS_PROCESS_STATE_WAITREAD;
							}
						} else {
							divert_to_dummy(p,-errno);
						}
					free(ufds);
					} else {
						divert_to_dummy(p,-ENOMEM);
					}
				}
			}
		break;

		/*
		Next, we have operations that do not modify any files
		in particular, but change the state of the file table
		within the process in question.
		*/

		case SYSCALL_chdir:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_chdir(path);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_getcwd:
			if(entering) {
				if(pfs_getcwd(path,sizeof(path))) {
					p->syscall_result = strlen(path)+1;
					if(p->syscall_result>args[1]) {
						p->syscall_result = -ERANGE;
					} else {
						tracer_copy_out(p->tracer,path,(void*)args[0],p->syscall_result);
					}
				} else {
					p->syscall_result = -errno;
				}
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_dup:
			if(entering) {
				p->syscall_result = pfs_dup(args[0]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_dup2:
			if(entering) {
				p->syscall_result = pfs_dup2(args[0],args[1]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		/*
		Next we have all of the system calls that work on
		a file name, rather than an open file.  In most cases,
		we use the (fast) tracer_copy_in to fetch the file
		name, and then invoke the pfs_  XXX We should have
		some sort of bounds checking on the path name.
		*/

		case SYSCALL_stat:
			decode_stat(p,entering,SYSCALL_stat,args,0);
			break;
		case SYSCALL_stat64:
			decode_stat(p,entering,SYSCALL_stat,args,1);
			break;
		case SYSCALL_lstat:
			decode_stat(p,entering,SYSCALL_lstat,args,0);
			break;
		case SYSCALL_lstat64:
			decode_stat(p,entering,SYSCALL_lstat,args,1);
			break;
		case SYSCALL_statfs:
			decode_statfs(p,entering,SYSCALL_statfs,args);
			break;

		case SYSCALL_access:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_access(path,args[1]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_chmod:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_chmod(path,args[1]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_chown:
		#ifdef SYSCALL_chown32
		case SYSCALL_chown32:
		#endif
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_chown(path,args[1],args[2]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_lchown:
		#ifdef SYSCALL_lchown32
		case SYSCALL_lchown32:
		#endif
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_lchown(path,args[1],args[2]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		#ifdef SYSCALL_truncate
		case SYSCALL_truncate:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_truncate(path,args[1]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;
		#endif

		#ifdef SYSCALL_truncate64
		case SYSCALL_truncate64:
			if(entering) {
				long long offset = args[1]+(((long long)args[2])<<32);
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_truncate(path,offset);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;
		#endif

		case SYSCALL_unlink:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_unlink(path);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_rename:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				tracer_copy_in(p->tracer,path2,(void*)args[1],sizeof(path2));
				p->syscall_result = pfs_rename(path,path2);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_link:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				tracer_copy_in(p->tracer,path2,(void*)args[1],sizeof(path2));
				p->syscall_result = pfs_link(path,path2);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_symlink:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				tracer_copy_in(p->tracer,path2,(void*)args[1],sizeof(path2));
				p->syscall_result = pfs_symlink(path,path2);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_readlink:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_readlink(path,path2,sizeof(path2));
				if(p->syscall_result<0) {
					p->syscall_result = -errno;
				} else {
					tracer_copy_out(p->tracer,path2,(void*)args[1],p->syscall_result);
				}
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_mknod:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_mknod(path,args[1],args[2]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_mkdir:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_mkdir(path,args[1]);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		case SYSCALL_rmdir:
			if(entering) {
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				p->syscall_result = pfs_rmdir(path);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;

		/*
		Although pfs_table supports the high-level operations
		opendir/readdir/closedir, all we can get a hold of at
		this level is getdents, which works on an open file
		descriptor.  We copy dirents out one by one using fdreaddir,
		and transform them into the type expected by the kernel.
		If we overrun the available buffer, immediately seek
		the fd back to where it was before.
		*/

		case SYSCALL_getdents:
			if(entering) {
				int fd = args[0];
				char *uaddr = (char*) args[1];
				int length = args[2];
				int result = 0;

				struct dirent *d;
				struct pfs_kernel_dirent buf;

				errno = 0;
				while((d=pfs_fdreaddir(fd))) {
					COPY_DIRENT(*d,buf);
					if(DIRENT_SIZE(buf)>length) {
						pfs_lseek(fd,d->d_off,SEEK_SET);
						errno = EINVAL;
						break;
					}
					tracer_copy_out(p->tracer,&buf,(void*)uaddr,buf.d_reclen);
					uaddr  += buf.d_reclen;
					length -= buf.d_reclen;
					result += buf.d_reclen;
				}

				if(result==0 && errno!=0) {
					p->syscall_result = -errno;
				} else {
					p->syscall_result = result;
				}
				divert_to_dummy(p,p->syscall_result);
			}
			break;


		#ifdef SYSCALL_getdents64
		case SYSCALL_getdents64:
			if(entering) {
				int fd = args[0];
				char *uaddr = (char*) args[1];
				int length = args[2];
				int result = 0;

				struct dirent *d;
				struct pfs_kernel_dirent64 buf;

				errno = 0;
				while((d=pfs_fdreaddir(fd))) {
					COPY_DIRENT(*d,buf);
					if(DIRENT_SIZE(buf)>length) {
						pfs_lseek(fd,d->d_off,SEEK_SET);
						errno = EINVAL;
						break;
					}
					tracer_copy_out(p->tracer,&buf,(void*)uaddr,buf.d_reclen);
					uaddr  += buf.d_reclen;
					length -= buf.d_reclen;
					result += buf.d_reclen;
				}

				if(result==0 && errno!=0) {
					p->syscall_result = -errno;
				} else {
					p->syscall_result = result;
				}
				divert_to_dummy(p,p->syscall_result);
			}
			break;
		#endif

		case SYSCALL_utime:
			if(entering) {
				struct utimbuf ut;
				tracer_copy_in(p->tracer,path,(void*)args[0],sizeof(path));
				if(args[1]) {
					tracer_copy_in(p->tracer,&ut,(void*)args[1],sizeof(ut));
				} else {
					ut.actime = ut.modtime = time(0);
				}
				p->syscall_result = pfs_utime(path,&ut);
				if(p->syscall_result<0) p->syscall_result = -errno;
				divert_to_dummy(p,p->syscall_result);
			}
			break;


		case SYSCALL_uname:
			if(pfs_false_uname) {
				struct utsname u;
				tracer_copy_in(p->tracer,&u,(void*)args[0],sizeof(struct utsname));
				strcpy(u.nodename,pfs_false_uname);
				tracer_copy_out(p->tracer,&u,(void*)args[0],sizeof(struct utsname));
			}
			break;

		/*
		Note that we call pfs_process_raise here so that the process data
		structures are made aware of the signal propagation, possibly kicking
		someone out of sleep.  However, we do *not* convert this call to
		a dummy, so that the sender can deliver itself, thus getting the
		correct data into the sa_info structure.
		*/

		case SYSCALL_kill:
		case SYSCALL_tkill:
			pfs_process_raise(args[0],args[1],0);
			break;

		/*
		We need to track the umask ourselves and use it in open.
		*/

		case SYSCALL_umask:
			if(entering) {
				int old_umask = current->umask;
				current->umask = args[0] & 0777;
				divert_to_dummy(p,old_umask);
			}
			break;

		/*
		The tracing mechanism re-parents traced children,
		so we must fake the parent pid if the child wants
		to send its parent a signal.
		*/

		case SYSCALL_getppid:
			divert_to_dummy(p,p->ppid);
			break;

		/*
		Always return the dummy uids.	
		*/

		case SYSCALL_geteuid32:
		case SYSCALL_geteuid:
		case SYSCALL_getuid32:
		case SYSCALL_getuid:
			divert_to_dummy(p,pfs_uid);
			break;

		case SYSCALL_getegid32:
		case SYSCALL_getegid:
		case SYSCALL_getgid32:
		case SYSCALL_getgid:
			divert_to_dummy(p,pfs_gid);
			break;

		case SYSCALL_getresuid32:
		case SYSCALL_getresuid:
			tracer_copy_out(p->tracer,&pfs_uid,(void*)args[0],sizeof(pfs_uid));
			tracer_copy_out(p->tracer,&pfs_uid,(void*)args[1],sizeof(pfs_uid));
			tracer_copy_out(p->tracer,&pfs_uid,(void*)args[2],sizeof(pfs_uid));
			divert_to_dummy(p,0);
			break;
		

		case SYSCALL_getresgid32:
		case SYSCALL_getresgid:
			tracer_copy_out(p->tracer,&pfs_gid,(void*)args[0],sizeof(pfs_uid));
			tracer_copy_out(p->tracer,&pfs_gid,(void*)args[1],sizeof(pfs_uid));
			tracer_copy_out(p->tracer,&pfs_gid,(void*)args[2],sizeof(pfs_uid));
			divert_to_dummy(p,0);
			break;

		case SYSCALL_setsid:
			if(entering) {
				current->tty[0] = 0;
			}
			break;

		/*
		Generally speaking, the kernel implements signal handling,
		so we just pass through operations such as sigaction and signal.
		However, we must keep track of which signals are allowed to
		interrupt I/O operations in progress.  Each process has an
		array, signal_interruptible, that records this. The SA_RESTART
		flag to sigaction can turn this on or off.  The traditional
		BSD signal() always turns it on.
		*/

		case SYSCALL_sigaction:
		case SYSCALL_rt_sigaction:
			if(entering) {
				int sig = args[0];
				struct pfs_kernel_sigaction act;
				tracer_copy_in(p->tracer,&act,(void*)args[1],sizeof(act));
				if(act.sa_flags&SA_RESTART) {
					current->signal_interruptible[sig] = 0;
				} else {
					current->signal_interruptible[sig] = 1;
				}
			}
			break;

		case SYSCALL_signal:
			if(entering) {
				int sig = args[0];
				current->signal_interruptible[sig] = 0;
			}
			break;

		/*
		Changing the userid is not allow, but for completeness,
		you can always change to your own uid.
		*/
		case SYSCALL_setgid:
		case SYSCALL_setgid32:
		case SYSCALL_setregid32:
		case SYSCALL_setregid:
		case SYSCALL_setuid:
		case SYSCALL_setuid32:
		case SYSCALL_setresuid32:
		case SYSCALL_setresuid:
		case SYSCALL_setresgid32:
		case SYSCALL_setresgid:
		case SYSCALL_setreuid32:
		case SYSCALL_setreuid:

			divert_to_dummy(p,0);
			break;

		/*
		These things are not currently permitted.
		*/
		case SYSCALL_chroot:
		case SYSCALL_mount:
		case SYSCALL_stime:
		case SYSCALL_sysfs:
		case SYSCALL_umount2:
		case SYSCALL_umount:
		case SYSCALL_uselib:
		case SYSCALL_lookup_dcookie:
		case SYSCALL_sys_epoll_create:
		case SYSCALL_sys_epoll_ctl:
		case SYSCALL_sys_epoll_wait:
		case SYSCALL_remap_file_pages:

			divert_to_dummy(p,-EPERM);
			break;
		/*
		These system calls are simply historical artifacts.
		*/
		case SYSCALL_acct:
		case SYSCALL_break:
		case SYSCALL_ftime:
		case SYSCALL_gtty:
		case SYSCALL_lock:
		case SYSCALL_mpx:
		case SYSCALL_profil:
		case SYSCALL_stty:
		case SYSCALL_ulimit:
			divert_to_dummy(p,-ENOSYS);
			break;

		/*
		A wide variety of calls have no relation to file
		access, so we simply send them along to the
		underlying OS.
		*/

		case SYSCALL__sysctl:
		case SYSCALL_adjtimex:
		case SYSCALL_afs_syscall:
		case SYSCALL_alarm:
		case SYSCALL_bdflush:
		case SYSCALL_brk:
		case SYSCALL_capget:
		case SYSCALL_capset:
		case SYSCALL_create_module:		
		case SYSCALL_delete_module:		
		case SYSCALL_get_kernel_syms:
		case SYSCALL_getgroups32:
		case SYSCALL_getgroups:
		case SYSCALL_getitimer:
		case SYSCALL_getpgid:
		case SYSCALL_getpgrp:
		case SYSCALL_getpid:
		case SYSCALL_getpriority:
		case SYSCALL_getrlimit:
		case SYSCALL_getrusage:
		case SYSCALL_getsid:
		case SYSCALL_gettid:
		case SYSCALL_gettimeofday:
		case SYSCALL_idle:
		case SYSCALL_init_module:		
		case SYSCALL_ioperm:
		case SYSCALL_iopl:
		case SYSCALL_ipc:
		case SYSCALL_madvise:
		case SYSCALL_mincore:
		case SYSCALL_mlock:
		case SYSCALL_mlockall:
		case SYSCALL_modify_ldt:
		case SYSCALL_mprotect:
		case SYSCALL_mremap:
		case SYSCALL_msync:
		case SYSCALL_munlock:
		case SYSCALL_munlockall:
		case SYSCALL_nanosleep:
		case SYSCALL_nice:
		case SYSCALL_pause:
		case SYSCALL_prctl:
		case SYSCALL_query_module:
		case SYSCALL_quotactl:
		case SYSCALL_reboot:
		case SYSCALL_rt_sigpending:
		case SYSCALL_rt_sigprocmask:
		case SYSCALL_rt_sigqueueinfo:
		case SYSCALL_rt_sigreturn:
		case SYSCALL_rt_sigsuspend:
		case SYSCALL_rt_sigtimedwait:
		case SYSCALL_sched_get_priority_max:
		case SYSCALL_sched_get_priority_min:
		case SYSCALL_sched_getparam:
		case SYSCALL_sched_getscheduler:
		case SYSCALL_sched_rr_get_interval:
		case SYSCALL_sched_setparam:
		case SYSCALL_sched_setscheduler:
		case SYSCALL_sched_yield:
		case SYSCALL_setdomainname:
		case SYSCALL_setfsgid32:
		case SYSCALL_setfsuid32:
		case SYSCALL_setgroups32:
		case SYSCALL_setgroups:
		case SYSCALL_sethostname:
		case SYSCALL_setitimer:
		case SYSCALL_setpgid:
		case SYSCALL_setpriority:
		case SYSCALL_setrlimit:
		case SYSCALL_settimeofday:
		case SYSCALL_set_tid_address:
		case SYSCALL_sgetmask:
		case SYSCALL_sigaltstack:
		case SYSCALL_sigpending:
		case SYSCALL_sigprocmask:
		case SYSCALL_sigreturn:
		case SYSCALL_sigsuspend:
		case SYSCALL_ssetmask:
		case SYSCALL_swapoff:
		case SYSCALL_swapon:
		case SYSCALL_sync:
		case SYSCALL_sysinfo:
		case SYSCALL_syslog:
		case SYSCALL_time:
		case SYSCALL_times:
		case SYSCALL_ugetrlimit:
		case SYSCALL_ustat:
		case SYSCALL_vhangup:
		case SYSCALL_vm86:
		case SYSCALL_vm86old:
		case SYSCALL_sched_setaffinity:
		case SYSCALL_sched_getaffinity:
		case SYSCALL_set_thread_area:
		case SYSCALL_get_thread_area:
		case SYSCALL_alloc_hugepages:
		case SYSCALL_free_hugepages:
		case SYSCALL_futex:
			break;

		/*
		These system calls get and set extended
		attributes on file systems.  Not only are
		they non-standard, they are only implemented
		by the jfs filesystem on Linux.  Libraries
		expect these to return EOPNOTSUPP.
		*/
		
                case SYSCALL_fgetxattr:
                case SYSCALL_flistxattr:
                case SYSCALL_fremovexattr:
                case SYSCALL_fsetxattr:
                case SYSCALL_getxattr:
                case SYSCALL_lgetxattr:
                case SYSCALL_listxattr:
                case SYSCALL_llistxattr:
                case SYSCALL_lremovexattr:
                case SYSCALL_lsetxattr:
		case SYSCALL_removexattr:
                case SYSCALL_setxattr:
                        divert_to_dummy(p,-EOPNOTSUPP);
                        break;

		/*
		These system calls could concievably be supported,
		but we haven't had the need or the time to attack
		them.
		*/

		case SYSCALL_getpmsg:
		case SYSCALL_io_cancel:
		case SYSCALL_io_destroy:
		case SYSCALL_io_getevents:
		case SYSCALL_io_setup:
		case SYSCALL_io_submit:
		case SYSCALL_ptrace:
		case SYSCALL_putpmsg:
		case SYSCALL_readahead:
		case SYSCALL_security:
		case SYSCALL_sendfile64:
		case SYSCALL_sendfile:
			/* fallthrough */

		/*
		If anything else escaped our attention, we must know
		about it in an obvious way.
		*/

		default:
			if(entering) {
				debug(D_SYSCALL,"warning: system call %d (%s) not supported",p->syscall,tracer_syscall_name(p->tracer,p->syscall));
				divert_to_dummy(p,-ENOSYS);
			}
			break;
	}

	if(!entering && p->state==PFS_PROCESS_STATE_KERNEL) {
		p->state = PFS_PROCESS_STATE_USER;
		if(p->syscall_args_changed) {
			tracer_args_set(p->tracer,p->syscall,p->syscall_args,TRACER_ARGS_MAX);
			p->syscall_args_changed = 0;
		}
		if(p->syscall_dummy) {
			tracer_result_set(p->tracer,p->syscall_result);
			p->syscall_dummy = 0;
			debug(D_SYSCALL,"= %d %s",p->syscall_result,p->syscall_result<0 ? strerror(-p->syscall_result) : "" );
		} else {
			debug(D_SYSCALL,"= ");
		}
	}

}

/*
Note that we clear the interrupted flag whenever
we start a new system call or leave an old one.
We don't want one system call to be interrupted
by a signal from a previous system call.
*/

void pfs_dispatch( struct pfs_process *p, int signum )
{
	struct pfs_process *oldcurrent = current;
	current = p;

	switch(p->state) {
		case PFS_PROCESS_STATE_KERNEL:
		case PFS_PROCESS_STATE_WAITWRITE:
			decode_syscall(p,0);
			break;
		case PFS_PROCESS_STATE_USER:
			p->interrupted = 0;
		case PFS_PROCESS_STATE_WAITREAD:
			decode_syscall(p,1);
			break;
		case PFS_PROCESS_STATE_WAITPID:
		case PFS_PROCESS_STATE_DONE:
			p->interrupted = 0;
			break;
		default:
			debug(D_PROCESS,"process %d in unexpected state %d",p->pid,p->state);
			break;
	}

	switch(p->state) {
		case PFS_PROCESS_STATE_KERNEL:
		case PFS_PROCESS_STATE_USER:
			tracer_continue(p->tracer,signum);
			break;
		case PFS_PROCESS_STATE_WAITPID:
		case PFS_PROCESS_STATE_WAITREAD:
		case PFS_PROCESS_STATE_WAITWRITE:
		case PFS_PROCESS_STATE_DONE:
			break;
		default:
			debug(D_PROCESS,"process %d in unexpected state %d",p->pid,p->state);
			break;
	}

	current = oldcurrent;
}
