/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "pfs_service.h"

extern "C" {
#include "debug.h"
#include "stringtools.h"
#include "domain_name.h"
#include "link.h"
#include "file_cache.h"
#include "full_io.h"
#include "http_query.h"
#include "hash_table.h"
#include "xmalloc.h"
#include "macros.h"
#include "sha1.h"
}

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/statfs.h>

#define HTTPFS_LINE_MAX 4096
#define HTTPFS_PORT 80

extern int pfs_master_timeout;
extern int pfs_checksum_files;
extern void pfs_abort();
extern const char * pfs_root_checksum;

/*
General theory of operation:

Each directory in an HTTPFS filesystem has a .parrothttpfsdir
file that lists the files in the directory, along with their metadata.
To list a directory, we must fetch the listing and go through each entry.
To stat() a file, we must look it up in the parent directory listing.
To open() a file, we must first look it up in the directory listing
to be sure that it exists, and then open it directly.  Note that we
cannot open a file without checking the directory first, because many
web servers will return "helpful" documents when a file or directory
is not found.

To avoid loading and re-loading the same metadata over and over again,
we cache loaded directory objects in a hash table.  We assume that
the filesystem is read only, so we simply keep loading them and never
clean up the cache.
*/

static struct hash_table * httpfs_cache = 0;

struct httpfs_dir {
	int result;
	int errnum;
	struct httpfs_dirent *list;
	char checksum[SHA1_DIGEST_ASCII_LENGTH];
};

struct httpfs_dirent {
	const char *name;
	struct pfs_stat info;
	const char *lname;
	struct pfs_stat linfo;
	char checksum[SHA1_DIGEST_ASCII_LENGTH];
	struct httpfs_dirent *next;
};

struct link * httpfs_openfile( pfs_name *name, const char *method, time_t stoptime )
{
	char url[PFS_PATH_MAX];
	sprintf(url,"http://%s:%d%s",name->host,name->port,name->rest);
	return http_query(url,method,stoptime);
}

struct httpfs_dir * httpfs_getdir( const char *url )
{
	char line[HTTPFS_LINE_MAX];
	struct httpfs_dir *dir = (struct httpfs_dir*) xxmalloc(sizeof(*dir));
	sha1_context_t context;
	unsigned char digest[SHA1_DIGEST_LENGTH];

	struct link * link = http_query(url,"GET",time(0)+pfs_master_timeout);
	if(!link) {
		dir->result = -1;
		dir->errnum = errno;
		dir->list = 0;
		dir->checksum[0] = 0;
		return dir;
	}

	dir->result = 0;
	dir->errnum = 0;
	dir->list = 0;

	if(pfs_root_checksum) {
		sha1_init(&context);
	}

	while(link_readline(link,line,sizeof(line),time(0)+pfs_master_timeout)) {
		struct httpfs_dirent *d = (struct httpfs_dirent *) xxmalloc(sizeof(*d));
		char name[PFS_PATH_MAX];
		char lname[PFS_PATH_MAX];

		if(pfs_root_checksum) {
			strcat(line,"\n");
			sha1_update(&context,(unsigned char *)line,strlen(line));
		}

		int fields = sscanf(line,"%[^\t]\t%lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %ld %ld %ld %s %[^\t]\t%lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %ld %ld %ld",
			name,
			&d->info.st_dev,
			&d->info.st_ino,
			&d->info.st_mode,
			&d->info.st_nlink,
			&d->info.st_uid,
			&d->info.st_gid,
			&d->info.st_rdev,
			&d->info.st_size,
			&d->info.st_blksize,
			&d->info.st_blocks,
			&d->info.st_atime,
			&d->info.st_mtime,
			&d->info.st_ctime,
			d->checksum,
			lname,
			&d->linfo.st_dev,
			&d->linfo.st_ino,
			&d->linfo.st_mode,
			&d->linfo.st_nlink,
			&d->linfo.st_uid,
			&d->linfo.st_gid,
			&d->linfo.st_rdev,
			&d->linfo.st_size,
			&d->linfo.st_blksize,
			&d->linfo.st_blocks,
			&d->linfo.st_atime,
			&d->linfo.st_mtime,
			&d->linfo.st_ctime);

		if(fields>=14) {
			if(fields==14) {
				lname[0] = 0;
			}
			d->name = xstrdup(name);
			d->lname = xstrdup(lname);
			d->next = dir->list;
			dir->list = d;
			continue;
		} else {
			break;
		}
	}

	link_close(link);

	if(pfs_root_checksum) {
		sha1_final(digest,&context);
		strcpy(dir->checksum,sha1_string(digest));
		debug(D_HTTP,"directory checksum: %s",dir->checksum);
	} else {
		dir->checksum[0] = 0;
	}

	return dir;

}

int httpfs_check_path_integrity( const char *urlin, const char *checksum );

struct httpfs_dir * httpfs_getdir_cached( const char *url )
{
	struct httpfs_dir *dir;
	char dirurl[PFS_PATH_MAX];

	sprintf(dirurl,"%s/.parrothttpfsdir",url);

	if(!httpfs_cache) httpfs_cache = hash_table_create(0,0);

	dir = (struct httpfs_dir *) hash_table_lookup(httpfs_cache,dirurl);
	if(dir) return dir;

	dir = httpfs_getdir(dirurl);

	if(dir && dir->result==0 && pfs_root_checksum) {
		httpfs_check_path_integrity(url,dir->checksum);
	}

	hash_table_insert(httpfs_cache,dirurl,dir,(void**)0);

	return dir;
}

int httpfs_check_path_integrity( const char *urlin, const char *checksum )
{
	char url[PFS_PATH_MAX];
	char dirname[PFS_PATH_MAX];
	const char *basename;
	struct httpfs_dir *dir;
	struct httpfs_dirent *d;
	int passed;

	if(!pfs_root_checksum) return 1;

	strcpy(url,urlin);

	while(1) {
		if(!strcmp(checksum,pfs_root_checksum)) {
			debug(D_HTTP,"%s has the root checksum",url);
			return 1;
		}

		string_dirname(url,dirname);
		basename = string_basename(url);

		dir = httpfs_getdir_cached(dirname);
		if(!dir) {
			debug(D_NOTICE,"couldn't get checksum for %s",dirname);
			pfs_abort();
		}

		passed = 0;

		for(d=dir->list;d;d=d->next) {
			if(!strcmp(d->name,basename)) {
				if(!strcmp(d->checksum,checksum)) {
					passed = 1;
					break;
				} else {
					debug(D_NOTICE,"checksum for %s was %s but should be %s",url,d->checksum,checksum);
					pfs_abort();
				}
			}
		}

		if(passed) {
			debug(D_HTTP,"directory checksum: %s",checksum);
			strcpy(url,dirname);
			checksum = dir->checksum;
		} else {
			debug(D_NOTICE,"directory checksum failed on %s",url);
			pfs_abort();
		}
	}
	return 1;
}

struct httpfs_dir * httpfs_getdir_byname( pfs_name *name, int do_parent )
{
	char url[PFS_PATH_MAX];
	struct httpfs_dir *dir;

	if(do_parent) {
		char dirname[PFS_PATH_MAX];
		string_dirname(name->rest,dirname);
		sprintf(url,"http://%s:%d%s",name->host,name->port,dirname);
	} else {
		sprintf(url,"http://%s:%d%s",name->host,name->port,name->rest);
	}

	dir = httpfs_getdir_cached(url);

	return dir;
}

int httpfs_search_dirent( pfs_name *name, const char *basename, int do_parent, struct httpfs_dirent *dresult )
{
	struct httpfs_dir *dir;
	struct httpfs_dirent *d;

	dir = httpfs_getdir_byname(name,do_parent);
	if(dir->result!=0) {
		errno = dir->errnum;
		return -1;
	}

	for(d=dir->list;d;d=d->next) {
		if(!strcmp(d->name,basename)) {
			*dresult = *d;
			return 0;
		}
	}

	errno = ENOENT;
	return -2;
}

int httpfs_get_dirent( pfs_name *name, struct httpfs_dirent *d )
{
	int result;
	const char *basename = string_basename(name->rest);

	result = httpfs_search_dirent(name,basename,1,d);
	if(result==-1 && errno==ENOENT) {
		// the directory list itself is not there
		result = httpfs_search_dirent(name,".",0,d);
	} else if(result==-2) {
		// the directory list is there, but no entry.
		errno = ENOENT;
		result = -1;
	}
	
	return result;
}

class pfs_file_httpfs : public pfs_file
{
private:
	struct link *link;
	pfs_stat info;
	sha1_context_t context;

public:
	pfs_file_httpfs( pfs_name *n, struct link *l, pfs_stat i ) : pfs_file(n) {
		link = l;
		info = i;
		if(pfs_checksum_files) {
			sha1_init(&context);
		}
	}

	virtual int close() {
		link_close(link);
		if(pfs_checksum_files) {
			struct httpfs_dirent d;
			if(httpfs_get_dirent(&name,&d)==0) {
				if(!strcmp(d.checksum,"0")) {
					debug(D_HTTP,"file checksum: not available");
				} else {
					unsigned char digest[SHA1_DIGEST_LENGTH];
					sha1_final(digest,&context);
					if(!strcmp(sha1_string(digest),d.checksum)) {
						debug(D_HTTP,"file checksum: %s",d.checksum);
					} else {
						debug(D_NOTICE,"file checksum for %s was %s should be %s",name.path,sha1_string(digest),d.checksum);
						pfs_abort();
					}
				}
			}
		}
		return 0;
	}

	virtual pfs_ssize_t read( void *d, pfs_size_t length, pfs_off_t offset ) {
		pfs_ssize_t actual;
		actual = link_read(link,(char*)d,length,LINK_FOREVER);
		if(pfs_checksum_files && actual>0) sha1_update(&context,(unsigned char *)d,actual);
		return actual;
	}

	virtual int fstat( struct pfs_stat *i ) {
		*i = info;
		return 0;
	}

	virtual pfs_ssize_t get_size() {
		return info.st_size;
	}

};

class pfs_service_httpfs : public pfs_service {
public:
	virtual int get_default_port() {
		return HTTPFS_PORT;
	}

	virtual pfs_file * open( pfs_name *name, int flags, mode_t mode ) {
		struct link *link;
		struct pfs_stat info;

		if((flags&O_ACCMODE)!=O_RDONLY) {
			errno = EROFS;
			return 0;
		}

		if(this->stat(name,&info)==0) {
			if(S_ISDIR(info.st_mode)) {
				errno = EISDIR;
				return 0;
			} else {
				link = httpfs_openfile(name,"GET",time(0)+pfs_master_timeout);
				if(link) {
					return new pfs_file_httpfs(name,link,info);
				} else {
					return 0;
				}
			}
		} else {
			return 0;
		}
	}

	pfs_dir * getdir( pfs_name *name ) {
		struct httpfs_dir *dir;
		struct httpfs_dirent *d;

		dir = httpfs_getdir_byname(name,0);
		if(dir->result!=0) {
			errno = dir->errnum;
			return 0;
		}

		pfs_dir *pdir = new pfs_dir(name);

		for(d=dir->list;d;d=d->next) {
			pdir->append(d->name);
		}

		return pdir;
	}

	virtual int lstat( pfs_name *name, struct pfs_stat *info ) {
		struct httpfs_dirent d;
		if(httpfs_get_dirent(name,&d)==0) {
			*info = d.info;
			return 0;
		} else {
			return -1;
		}
	}

	virtual int stat( pfs_name *name, struct pfs_stat *info ) {
		struct httpfs_dirent d;
		if(httpfs_get_dirent(name,&d)==0) {
			if(S_ISLNK(d.info.st_mode)) {
				*info = d.linfo;
			} else {
				*info = d.info;
			}
			return 0;
		} else {
			return -1;
		}

		errno = ELOOP;
		return -1;
	}

	virtual int unlink( pfs_name *name ) {
		errno = EROFS;
		return -1;
	}

	virtual int access( pfs_name *name, mode_t mode ) {
		struct pfs_stat info;
		if(this->stat(name,&info)==0) {
			if(mode&W_OK) {
				errno = EROFS;
				return -1;
			} else {
				return 0;
			}
		} else {
			return -1;
		}
	}

	virtual int chmod( pfs_name *name, mode_t mode ) {
		errno = EROFS;
		return -1;
	}

	virtual int chown( pfs_name *name, uid_t uid, gid_t gid ) {
		errno = EROFS;
		return -1;
	}

	virtual int lchown( pfs_name *name, uid_t uid, gid_t gid ) {
		errno = EROFS;
		return -1;
	}

	virtual int truncate( pfs_name *name, pfs_off_t length ) {
		errno = EROFS;
		return -1;
	}

	virtual int utime( pfs_name *name, struct utimbuf *buf ) {
		errno = EROFS;
		return -1;
	}

	virtual int rename( pfs_name *oldname, pfs_name *newname ) {
		errno = EROFS;
		return -1;
	}

	virtual int chdir( pfs_name *name, char *newpath ) {
		struct pfs_stat info;
		if(this->stat(name,&info)==0) {
			if(S_ISDIR(info.st_mode)) {
				return 0;
			} else {
				errno = ENOTDIR;
				return -1;
			}
		} else {
			return -1;
		}
	}

	virtual int link( pfs_name *oldname, pfs_name *newname ) {
		errno = EROFS;
		return -1;
	}

	virtual int symlink( const char *linkname, pfs_name *newname ) {
		errno = EROFS;
		return -1;
	}

	virtual int readlink( pfs_name *name, char *buf, pfs_size_t bufsiz ) {
		struct httpfs_dirent d;

		if(httpfs_get_dirent(name,&d)==0) {
			if(S_ISLNK(d.info.st_mode)) {
				int length;
				strncpy(buf,d.lname,bufsiz);
				length = MIN(bufsiz,strlen(d.lname));
				buf[length] = 0;
				return length;
			} else {
				errno = EINVAL;
				return -1;
			}
		} else {
			return -1;
		}
	}

	virtual int mkdir( pfs_name *name, mode_t mode ) {
		errno = EROFS;
		return -1;
	}

	virtual int rmdir( pfs_name *name ) {
		errno = EROFS;
		return -1;
	}

};

static pfs_service_httpfs pfs_service_httpfs_instance;
pfs_service *pfs_service_httpfs = &pfs_service_httpfs_instance;


